#! /usr/bin/wish
#
# same.tcl : SameGame 風のパズルゲーム「おとしてポン」
#
#            Copyright (C) 1998 - 2000, 2014, 2019 by Makoto Hiroi
#
#    global 変数
#        board       : 図形 id を格納する配列
#        piece       : 駒の種別を格納する配列
#                       -1 : 削除, 0 : 駒無し（背景色）, 1 以上が駒
#        data_table  : データを格納する配列
#        score_table : ハイスコアを格納する配列
#        data_num    : 問題番号
#        score       : 現在のスコア
#        display     : 表示用ラベル（配列）
#
#  （注意）
#    キャンバスの部品を削除することで表示を消すと、
#    部品の id をどんどん消費することになるので、
#    表示の消去は背景色で塗りつぶすことにする。
#    よって、部品の削除は行わない。
#

# ********** 色テーブル(0 は背景色) **********
set color_table(0) gray
set color_table(1) orange
set color_table(2) yellow
set color_table(3) green
set color_table(4) red
set color_table(5) blue

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/same.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
    }
}

# 駒の配置テーブルの初期化
proc init_piece_table {} {
    global board piece color_table
    # 初期化
    for {set x 0} {$x < 16} {incr x 1} {
	for {set y 0} {$y < 16} {incr y 1} {
	    set board($x,$y) 0
	    set piece($x,$y) 0
	}
    }
    # 円を描く
    for {set x 1} {$x < 15} {incr x 1} {
	for {set y 1} {$y < 15} {incr y 1} {
	    set x1 [expr 24 * ($x - 1) + 2]
	    set y1 [expr 24 * ($y - 1) + 2]
	    set x2 [expr $x1 + 24]
	    set y2 [expr $y1 + 24]
	    set board($x,$y) [.c0 create oval $x1 $y1 $x2 $y2 -outline $color_table(0)]
	}
    }
}

# 駒を消すことができるか
proc check_same_piece {x y c} {
    global piece
    set n [expr $x - 1]
    if {$piece($n,$y) == $c } { return 1 }
    set n [expr $x + 1]
    if {$piece($n,$y) == $c } { return 1 }
    set n [expr $y - 1]
    if {$piece($x,$n) == $c } { return 1 }
    set n [expr $y + 1]
    if {$piece($x,$n) == $c } { return 1 }
    return 0
}

# 終了チェック
proc check_finish {} {
    global piece
    for {set x 1} {$x < 15} {incr x 1} {
	for {set y 14} {$y > 0} {incr y -1} {
            set c $piece($x,$y)            
	    if {$c > 0 && [check_same_piece $x $y $c] } {
                # 取れる駒がある
		return 0
	    } elseif {$c == 0} {
		break
	    }
	}
    }
    return 1
}

# ボーナスポイントを計算する
proc count_bonus {} {
    global piece
    set count_color 5
    for {set x 1} {$x <= 5} {incr x} {
	set count_color_flag($x) 0
    }
    for {set x 1} {$x < 15} {incr x 1} {
	for {set y 14} {$y > 0} {incr y -1} {
            set c $piece($x,$y)            
	    if {$c > 0 && $count_color_flag($c) == 0} {
		set count_color_flag($c) 1
		incr count_color -1
	    }
	}
    }
    if {$count_color == 5} {
	# 全部消したので１０万点のボーナス
	return 100000
    }
    # １種類につき１万点のボーナス
    return [expr $count_color * 10000]
}
    

# 駒を消す
proc delete_piece_sub {x y c} {
    global piece del_count
    incr del_count 1
    # -1 は駒を削除したことを表す
    set piece($x,$y) -1
    set n [expr $x - 1]
    if {$piece($n,$y) == $c } {
	delete_piece_sub $n $y $c
    }
    set n [expr $x + 1]
    if {$piece($n,$y) == $c } {
	delete_piece_sub $n $y $c
    }
    set n [expr $y - 1]
    if {$piece($x,$n) == $c } {
	delete_piece_sub $x $n $c
    }
    set n [expr $y + 1]
    if {$piece($x,$n) == $c } {
	delete_piece_sub $x $n $c
    }
}

# 色を変える
proc change_piece_color {x} {
    global piece board color_table
    for {set y 1} {$y <= 14} {incr y} {
	set c $piece($x,$y)
        .c0 itemconfigure $board($x,$y) -fill $color_table($c)
    }
}

# 駒を下に落とす
proc down_piece {} {
    global piece
    for {set x 1} {$x < 15} {incr x 1} {
        set flag 0
	for {set y 14} {$y > 0} {} {
	    if {$piece($x,$y) == -1} {
		# 下に一つずつ移動させる
		set flag 1
		for {set to $y} {$to > 0} {incr to -1} {
		    set from [expr $to - 1]
		    set piece($x,$to) $piece($x,$from)
		}
	    } else {
		incr y -1
	    }
	}
	if $flag {
	    change_piece_color $x
	}
    }
}

# 駒のある列を探す
proc search_piece {x} {
    global piece
    for {set x1 $x} {$x1 < 15} {incr x1} {
        # 一番下に駒があれば発見
	if { $piece($x1,14) > 0 } {
	    return $x1
	}
    }
    return 0
}

# 駒を横に移動
proc slide_piece {} {
    global piece
    for {set x 1} {$x < 15} {incr x} {
	if {$piece($x,14) == 0} {
	    # 駒がない列を発見
	    set x1 [search_piece [expr $x + 1]]
	    if {$x1 > 0} {
		for {set y 1} {$y < 15} {incr y} {
		    set piece($x,$y) $piece($x1,$y)
		    set piece($x1,$y) 0
		}
		change_piece_color $x
		change_piece_color $x1
	    }
	}
    }
}

# スコアのセーブ
proc write_score {} {
    global argv0 score_table
    # same.tcl 起動パスを取得する
    set path [file dirname $argv0]
    # スコアファイルをライトオープン
    set f [open "$path/same.sco" w]
    set max [array size score_table]
    for {set i 0} {$i < $max} {incr i} {
	puts $f $score_table($i)
    }
    close $f
}

# ハイスコアの確認
proc check_hi_score {} {
    global score score_table data_num
    # ボーナスポイントの計算
    incr score [count_bonus]
    if {$score > $score_table($data_num)} {
	set score_table($data_num) $score
        print_hi_score
	write_score
        set mes [format "おめでとう！ハイスコア %6d 点です" $score]
    } else {
	set mes [format "スコアは %6d 点でした" $score]
    }
    tk_messageBox -type ok -message $mes
}


# 駒を取り除く
proc delete_piece {x1 y1} {
    global piece del_count score
    set x [expr ($x1 - 2) / 24 + 1]
    set y [expr ($y1 - 2) / 24 + 1]
    set c $piece($x,$y)
    if { ($c > 0) && [check_same_piece $x $y $c] } {
	# 再帰で駒を消していく
	set del_count 0
	delete_piece_sub $x $y $c
    }
    down_piece
    slide_piece
    # スコアの計算
    incr del_count -1
    set score [expr $del_count * $del_count + $score]
    print_score
    if [check_finish] {
	# 終了
        check_hi_score
	next_game 1
    }
}

# 盤面を表示する
proc display_board {n} {
    global data_table board piece color_table
    set i 0
    for {set x 1} {$x < 15} {incr x 1} {
	for {set y 1} {$y < 15} {incr y 1} {
            set c [string index $data_table($n) $i]
	    set piece($x,$y) $c
	    .c0 itemconfigure $board($x,$y) -fill $color_table($c)
            incr i
	}
    }
}    

# データの読み込み
proc load_data {} {
    global argv0 data_table score_table
    # same.tcl 起動パスを取得する
    set path [file dirname $argv0]
    # データファイルをオープン
    set f [open "$path/same.dat" r]
    set i 0
    while {[gets $f line] >= 0} {
	set data_table($i) $line
	incr i
    }
    close $f
    # スコアファイルをオープン
    if [file exist "$path/same.sco"] {
	set f [open "$path/same.sco" r]
        set j 0
	while {[gets $f line] >= 0} {
	    set score_table($j) $line
	    incr j
	}
	close $f
    } else {
        # スコアファイルがない場合
	for {set j 0} {$j < $i} {incr j} {
	    set score_table($j) 0
	}
    }
}

# ハイスコアを表示する
proc print_hi_score {} {
    global display data_num score_table
    set display(hi_score) \
        [format "Stage %3d  Hi-Score %6d" [expr $data_num + 1] $score_table($data_num)]
}

# スコアを表示する
proc print_score {} {
    global score display
    set display(score) [format "Score %6d" $score]
}

# 次のゲームへ
proc next_game {n} {
    global data_table data_num score
    set score 0
    set max [array size data_table]
    incr data_num $n
    if {$data_num >= $max} {
	set data_num [expr $max - 1]
    } elseif {$data_num < 0 } {
	set data_num 0
    }
    display_board $data_num
    print_hi_score
    print_score
}

# ゲームの初期化
proc init_game {} {
    global data_num display score
    init_piece_table
    load_data
    set data_num 0
    set score 0
    next_game 0
}

# ********** 画面の生成 **********
canvas .c0 -width 340 -height 340 -bg $color_table(0)
wm title . "OtoshitePon"
# メニューの設定
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games" -under 0 -menu .m.m1
.m add command -label "Help"  -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Exit" -under 0 -command "exit"

option add *font "{Noto Sans Mono CJK JP} 12"

label .l0 -textvariable display(hi_score)
label .l1 -textvariable display(score)
frame .f0
button .b0 -text "+1"   -command "next_game 1"
button .b1 -text "+10"   -command "next_game 10"
button .b2 -text "+50"   -command "next_game 50"
button .b3 -text "-1"   -command "next_game -1"
button .b4 -text "-10"   -command "next_game -10"
button .b5 -text "-50"   -command "next_game -50"
button .b6 -text "Retry" -command "next_game 0"
pack .b0 .b1 .b2 .b3 .b4 .b5 .b6 -in .f0 -side left
pack .f0 .l0 .l1 .c0

# ********** バインドの設定 **********
bind .c0  <Button-1>  "delete_piece %x %y"

# ********** ゲームの初期化 **********
init_game

# ********** end of file **********
