#! /usr/bin/wish
#
# ten_上海.tcl : 上海風足して 10 になるカードを取り除くパズルゲーム
#
#                Copyright (C) 1999,2000,2014,2019 by Makoto Hiroi
#
#
# 外部変数
#   board()   : 盤面（データを格納）
#               上位 4 ビットで色、下位 4 ビットで数字を表す
#   number()  : テキスト ID を格納
#   numstr()  : 表示する文字列
#   color()   : 表示する色
#   piece()   : 図形 ID を格納
#   px,py,pz  : 選択した数字の座標( -1 は未選択 )
#   rest      : 残りの枚数
#   play_flag : ゲームの状態
#               0 : not play
#               1 : play 
#               2 : use search or takeback
#               4 : 検索中
#   buff1     : メッセージ表示用バッファ
#   buff2     : 名前入力用バッファ
#   time      : ゲーム開始時刻
#   id        : after コマンドが返す固有番号
#   history() : 取った牌の履歴
#   name()    : トップテン(1 - 10, 0 is dummy)
#   date()
#   score()
#

# 背景色
set bgcolor(0) gray90
set bgcolor(1) gray95
set bgcolor(2) white
set bgcolor(3) white

# 共通サブルーチンロード
source "ten_sub.tcl"

# 盤面から色を求める
proc get_color {x y z} {
    global board
    return [expr $board($x,$y,$z) / 16]
}

# 盤面から数字を求める
proc get_number {x y z} {
    global board
    return [expr $board($x,$y,$z) % 16]
}

# 牌をセレクトできるか
proc check_select {x y z} {
    global board
    # 上が空いていること
    if {$z == 3 || $board($x,$y,[expr $z + 1]) == 0} {
	# ４方向のいずれかが空いていること
	if {$x == 0 || $x == 5 || $y == 0 || $y == 5} {
	    return 1
	}
	if {$board([expr $x + 1],$y,$z) == 0} {
	    return 1
	}
	if {$board([expr $x - 1],$y,$z) == 0} {
	    return 1
	}
	if {$board($x,[expr $y + 1],$z) == 0} {
	    return 1
	}
	if {$board($x,[expr $y - 1],$z) == 0} {
	    return 1
	}
    }
    return 0
}

# 取れる牌があるか
proc search_piece {} {
    global board piece
    set buff ""
    set piece_list ""
    set result ""
    set len 0
    for {set x 0} {$x < 6} {incr x} {
	for {set y 0} {$y < 6} {incr y} {
	    for {set z 3} {$z >= 0} {incr z -1} {
		if {$board($x,$y,$z) > 0 && [check_select $x $y $z]} {
		    set buff [concat $buff $board($x,$y,$z)]
		    set piece_list [concat $piece_list $piece($x,$y,$z)]
		    incr len
		    break
		}
	    }
	}
    }
    for {set i 0} {$i < $len} {incr i} {
	for {set j [expr $i + 1]} {$j < $len} {incr j} {
	    set p1 [lindex $buff $i] 
	    set p2 [lindex $buff $j]
	    if {[expr $p1 /16 - $p2 / 16] == 0} {
		if {[expr $p1 % 16 + $p2 % 16] == 10} {
		    # 取れる牌があるよ
		    set result [concat $result \
			    [lindex $piece_list $i] [lindex $piece_list $j]]
		}
	    }
	}
    }
    return $result
}


# 盤面の初期化
proc make_board {} {
    global board
    # piece_table は局所変数
    for {set i 0; set c 0} {$c < 4} {incr c} {
	for {set n 1} {$n <= 9} {incr n} {
	    for {set j 0} {$j < 4} {incr j} {
		set piece_table($i) [expr $c * 16 + $n]
		incr i
	    }
	}
    }
    # 乱数でかき回す
    for {set j 0} {$j < $i} {incr j} {
	set n    [expr int( rand() * $i )]
	set temp $piece_table($n)
	set piece_table($n) $piece_table($j)
	set piece_table($j) $temp
    }
    # board にセット
    set i 0
    for {set z 0} {$z < 4} {incr z} {
	for {set y 0} {$y < 6} {incr y} {
	    for {set x 0} {$x < 6} {incr x} {
		set board($x,$y,$z) $piece_table($i)
		incr i
	    }
	}
    }
}

# 盤面の描画
proc draw_board {} {
    global number numstr piece color
    for {set z 0} {$z < 4} {incr z} {
	for {set y 0} {$y < 6} {incr y} {
	    for {set x 0} {$x < 6} {incr x} {
		set n [get_number $x $y $z]
		set c [get_color $x $y $z]
		.c0 itemconfigure $number($x,$y,$z) \
                        -text $numstr($n) \
			-fill $color($c)
		.c0 raise $piece($x,$y,$z)
		.c0 raise $number($x,$y,$z)
	    }
	}
    }
}

# 終了チェック
proc check_finish {} {
    global board rest time id buff2 play_flag
    if {$rest == 0} {
	# 終了
	set t [clock seconds]
	set s [expr $t - $time]
	after cancel $id
	set ranking [check_hi_score $s]
	if {$ranking > 0 && $play_flag == 1} {
	    input_hi_score_name $ranking 
	    tkwait window .t2
	    update_score $buff2 $t $s $ranking
	    write_score_file
	    open_score_window $ranking
	} else {
	    tk_messageBox -type ok \
		-message [format "おめでとう %s です" [change_seconds $s]]
	}
	return 1
    } else {
	# 手詰まりチェック
	if {[search_piece] == ""} {
	    after cancel $id
	    tk_messageBox -type ok -message "手詰まりです"
	    return 1
	}
    }
    return 0
}

# 牌を取り除く
proc remove_piece {x y z} {
    global piece number board rest bgcolor history
    .c0 lower $number($x,$y,$z)
    .c0 lower $piece($x,$y,$z)
    .c0 itemconfigure $piece($x,$y,$z) -fill $bgcolor($z)
    set history($rest) [concat $board($x,$y,$z) $x $y $z]
    set board($x,$y,$z) 0
    incr rest -1
}

# 同じ平面にある数字 ID を探す
proc get_number_id {z} {
    global board number
    for {set x 0} {$x < 6} {incr x} {
	for {set y 0} {$y < 6} {incr y} {
	    if {$board($x,$y,$z) != 0} {
		return $number($x,$y,$z)
	    }
	}
    }
    return 0
}

# 牌を戻す
proc takeback {} {
    global play_flag number piece board rest history px py pz
    if {$play_flag != 0 && $px != -1} {
	# 牌を選択している時はキャンセル扱いにする
	push_piece $px $py $pz
	return
    }
    if {!($play_flag & 0x03) || $rest == 144} return
    set play_flag 2
    for {set i 0} {$i < 2} {incr i} {
	incr rest
	set h $history($rest)
	set p [lindex $h 0]
	set x [lindex $h 1]
	set y [lindex $h 2]
	set z [lindex $h 3]
	set num_id [get_number_id $z]
	if {$num_id == 0} {
	    .c0 raise $piece($x,$y,$z)
	    .c0 raise $number($x,$y,$z)
	} else {
	    .c0 raise $piece($x,$y,$z)  $num_id
	    .c0 raise $number($x,$y,$z) $piece($x,$y,$z)
	}
	# get_number_id のサーチが終わってからセットすること
	set board($x,$y,$z) $p
    }
}

# 検索
proc search {} {
    global play_flag px py pz
    if {!($play_flag & 0x03)} return
    if {$px != -1} {
	# 選択していたらキャンセルして表示する
	push_piece $px $py $pz
    }
    set play_flag 4
    set pieces [search_piece]
    set len [llength $pieces]
    set i 0
    while {$i < $len} {
	set p1 [lindex $pieces $i]
	incr i
	set p2 [lindex $pieces $i]
	incr i
	set bg1 [.c0 itemcget $p1 -fill]
	set bg2 [.c0 itemcget $p2 -fill]
	.c0 itemconfigur $p1 -fill darkgray
	.c0 itemconfigur $p2 -fill darkgray
	update
	after 500
	.c0 itemconfigur $p1 -fill $bg1
	.c0 itemconfigur $p2 -fill $bg2
    }
    set play_flag 2
}


# バインディング
proc push_piece {x y z} {
    global board number piece play_flag px py pz bgcolor
    # 牌を選択できるか
    if {!($play_flag & 0x03) || [check_select $x $y $z] == 0} return
    if {$px == -1} {
	# 最初の選択
	set px $x
	set py $y
	set pz $z
	.c0 itemconfigure $piece($x,$y,$z) -fill darkgray
    } elseif {$px == $x && $py == $y && $pz == $z} {
	# 同じであればキャンセル扱いにする
	.c0 itemconfigure $piece($x,$y,$z) -fill $bgcolor($z)
	set px -1
	set py -1
	set pz -1
    } else {
	set n1 [get_number $px $py $pz]
	set c1 [get_color $px $py $pz]
	set n2 [get_number $x $y $z]
	set c2 [get_color $x $y $z]
	if {$c1 == $c2 && [expr $n1 + $n2] == 10} {
	    .c0 itemconfigure $piece($x,$y,$z) -fill darkgray
	    update
	    after 200
	    # 取り除く
	    remove_piece $px $py $pz
	    remove_piece $x $y $z
	    display_message
	    set px -1
	    set py -1
	    set pz -1
	    if [check_finish] {
		set play_flag 0
	    }
	}
    }
}

# メッセージの表示
proc display_message {} {
    global time buff1 rest
    set t [expr [clock seconds] - $time]
    set buff1 [format "残り %3d 枚    時間 %5s" $rest [change_seconds $t]]
}

# 時間の表示
proc display_time {} {
    global id
    display_message
    set id [after 1000 display_time]
}

# ゲームの開始
proc start_game {} {
    global play_flag time rest id px py pz
    if {$play_flag > 0} {
	after cancel $id
    }
    while 1 {
	make_board
	if {[search_piece] != ""} break
    }
    draw_board
    set rest 144
    set play_flag 1
    set time [clock seconds]
    set px -1
    set py -1
    set pz -1
    display_time
}


# メニューの設定
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games"    -under 0 -menu .m.m1
.m add command -label "Takeback" -under 0 -command "takeback"
.m add command -label "Search"   -under 0 -command "search"
.m add command -label "Help"     -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start"   -under 0 -command "start_game"
.m.m1 add command -label "HiScore" -under 0 -command "open_score_window 0"
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"

# オプションの設定
option add *font "{Noto Sans Mono CJK JP} 12"

# 画面の初期化
canvas .c0 -width 380 -height 380

# 背景
.c0 create rectangle 0 0 379 379 -fill darkgreen

# 牌の配置は３次元配列
set i 50
for {set z 0} {$z < 4} {incr z; incr i -5} {
    for {set y 0} {$y < 6} {incr y} {
	for {set x 0} {$x < 6} {incr x} {
	    #
	    set x1 [expr $x * 50 + $i]
	    set y1 [expr $y * 50 + $i]
	    set x2 [expr $x1 + 49]
	    set y2 [expr $y1 + 49]
	    set piece($x,$y,$z) \
		    [.c0 create rectangle $x1 $y1 $x2 $y2 -fill $bgcolor($z)]
	    set number($x,$y,$z) \
		    [.c0 create text [expr $x1 + 25] [expr $y1 + 25] \
		        -text "  " \
	                -font "{Noto Sans Mono CJK JP} 28 bold"]
	    .c0 bind $number($x,$y,$z) <Button-1> "push_piece $x $y $z"
	    .c0 bind $piece($x,$y,$z) <Button-1> "push_piece $x $y $z"
	}
    }
}

# 表示用ラベル
label .l1 -textvariable buff1 -bg darkgreen -fg white -anchor e

pack .l1 .c0 -fill x

# 窓の題名
wm title . "TEN"
wm resizable . 0 0

# 初期化
expr srand( [clock seconds] )
set play_flag 0
set path_name [file dirname $argv0]
set score_file "$path_name/TEN_上海.SCO"

# スコアファイルのリード
read_score_file
focus -force .

# end of file

